#include <windows.h>
#include <math.h>
#include <stdlib.h>
#include <stdio.h>


//-------------------------------------
//  MersenneTwister
//-------------------------------------
#define MT_IA  397
#define MT_LEN 624

class MersenneTwister
{
    unsigned long m_buffer[MT_LEN];
    int m_index;

public:
    MersenneTwister();
    unsigned long integer();
    float poissonInterval(float ooLambda) { return -logf(1.0f - integer() * 2.3283e-10f) * ooLambda; }
};

MersenneTwister::MersenneTwister()
{
    for (int i = 0; i < MT_LEN; i++)
        m_buffer[i] = rand();
    m_index = 0;
    for (int i = 0; i < MT_LEN * 100; i++)
        integer();
}

unsigned long MersenneTwister::integer()
{
    // Indices
    int i = m_index;
    int i2 = m_index + 1; if (i2 >= MT_LEN) i2 = 0; // wrap-around
    int j = m_index + MT_IA; if (j >= MT_LEN) j -= MT_LEN; // wrap-around

    // Twist
    unsigned long s = (m_buffer[i] & 0x80000000) | (m_buffer[i2] & 0x7fffffff);
    unsigned long r = m_buffer[j] ^ (s >> 1) ^ ((s & 1) * 0x9908B0DF);
    m_buffer[m_index] = r;
    m_index = i2;

    // Swizzle
    r ^= (r >> 11);
    r ^= (r << 7) & 0x9d2c5680UL;
    r ^= (r << 15) & 0xefc60000UL;
    r ^= (r >> 18);
    return r;
}


//-------------------------------------
//  Global & thread state
//-------------------------------------
static const int kMaxThreads = 4;       // This test is hardcoded for Quad-Core Xeon

struct ThreadStats
{
    LONGLONG workDone;
    LONGLONG iterations;
    LONGLONG overshoot;
};

struct GlobalState
{
    CRITICAL_SECTION criticalSection;
    HANDLE initSemaphore;
    HANDLE startEvent;
    float ooFreq;
    float secsPerWorkUnit;
    float timeLimit;
    float averageUnlockedCount;
    float averageLockedCount;
    ThreadStats threadStats[kMaxThreads];
};

GlobalState globalState = { 0 };

struct BenchmarkParams
{
    int threadCount;
    float lockInterval;     // In seconds
};

BenchmarkParams benchmarkParams[] =
{
    // Reference
    1, 10e-3f,      // 10 ms        100/s

    // Test 15000 locks per second with multiple threads
    1, 1/15000.0f,
    2, 1/15000.0f,
    3, 1/15000.0f,
    4, 1/15000.0f,

    // Test various lock rates with 2 threads
    2, 10e-9f,      // 10 ns        100000000/s
    2, 31.6e-9f,    // 31.6 ns      31600000/s
    2, 100e-9f,     // 100 ns       10000000/s
    2, 316e-9f,     // 316 ns       3160000/s
    2, 1e-6f,       // 1 us         1000000/s
    2, 3.16e-6f,    // 3.16 us      316000/s
    2, 10e-6f,      // 10 us        100000/s
    2, 31.6e-6f,    // 31.6 us      31600/s
    2, 100e-6f,     // 100 us       10000/s
};



//-------------------------------------
//  ThreadProc
//-------------------------------------
DWORD WINAPI ThreadProc(LPVOID lpParameter)
{
    // Initialize
    MersenneTwister random;
    LARGE_INTEGER start, end;
    ThreadStats threadStats = { 0 };
    float elapsedTime = 0;
    int workunits = 0;

    // Indicate ready, wait for start signal
    ReleaseSemaphore(globalState.initSemaphore, 1, NULL);
    WaitForSingleObject(globalState.startEvent, INFINITE);

    // Go
    QueryPerformanceCounter(&start);
    for (;;)
    {
        // Do some work without holding the lock
        workunits = (int) (random.poissonInterval(globalState.averageUnlockedCount) + 0.5f);
        for (int i = 1; i < workunits; i++)
            random.integer();       // Do one work unit
        threadStats.workDone += workunits;

        QueryPerformanceCounter(&end);
        elapsedTime = (end.QuadPart - start.QuadPart) * globalState.ooFreq;
        if (elapsedTime >= globalState.timeLimit)
            break;

        // Do some work while holding the lock
        EnterCriticalSection(&globalState.criticalSection);
        workunits = (int) (random.poissonInterval(globalState.averageLockedCount) + 0.5f);
        for (int i = 1; i < workunits; i++)
            random.integer();       // Do one work unit
        threadStats.workDone += workunits;
        LeaveCriticalSection(&globalState.criticalSection);

        threadStats.iterations++;
        QueryPerformanceCounter(&end);
        elapsedTime = (end.QuadPart - start.QuadPart) * globalState.ooFreq;
        if (elapsedTime >= globalState.timeLimit)
            break;
    }

    // Estimated the number of work units which went over the time limit:
    threadStats.overshoot = min(workunits, (int) ((elapsedTime - globalState.timeLimit) / globalState.secsPerWorkUnit));

    int threadNumber = (int) lpParameter;
    globalState.threadStats[threadNumber] = threadStats;
    return 0;
}


//-------------------------------------
//  CalcSecsPerWorkUnit
//-------------------------------------
float CalcSecsPerWorkUnit()
{
    MersenneTwister random;
    LARGE_INTEGER start, end;
    int count = 100000000;

    QueryPerformanceCounter(&start);
    for (int i = 0; i < count; i++)
        random.integer();
    QueryPerformanceCounter(&end);

    float elapsedTime = (end.QuadPart - start.QuadPart) * globalState.ooFreq;
    return elapsedTime / count;
}


//-------------------------------------
//  main
//-------------------------------------
int main()
{
    // Initialize
    SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_ABOVE_NORMAL);    // This is the watchdog thread
    InitializeCriticalSection(&globalState.criticalSection);
    globalState.initSemaphore = CreateSemaphore(NULL, 0, kMaxThreads, NULL);
    globalState.startEvent = CreateEvent(NULL, TRUE, FALSE, NULL);
    globalState.timeLimit = 1.0f;
    LARGE_INTEGER freq;
    QueryPerformanceFrequency(&freq);
    globalState.ooFreq = 1.0f / freq.QuadPart;
    globalState.secsPerWorkUnit = CalcSecsPerWorkUnit();
    printf("secsPerWorkUnit=%e\n", globalState.secsPerWorkUnit);
    //globalState.secsPerWorkUnit = 6.349e-9f;

    for (int b = 0; b < sizeof(benchmarkParams) / sizeof(benchmarkParams[0]); b++)
    {
        float avgWorkUnitsBetweenLocks = benchmarkParams[b].lockInterval / globalState.secsPerWorkUnit;
        int threadCount = benchmarkParams[b].threadCount;
        static const int steps = 200;
        for (int s = 0; s <= steps; s++)
        {
            globalState.averageLockedCount = avgWorkUnitsBetweenLocks * s / steps;
            globalState.averageUnlockedCount = avgWorkUnitsBetweenLocks * (steps - s) / steps;

            // Spawn threads, wait for all ready
            ResetEvent(globalState.startEvent);
            HANDLE threads[kMaxThreads];
            for (int t = 0; t < threadCount; t++)
            {
                threads[t] = CreateThread(NULL, 0, ThreadProc, (LPVOID) t, 0, NULL);
                SetThreadAffinityMask(threads[t], 1 << (t * 2));    // Hardcoded for Quad-Core Xeon
            }
            for (int t = 0; t < threadCount; t++)
                WaitForSingleObject(globalState.initSemaphore, INFINITE);

            // Start & end threads
            SetEvent(globalState.startEvent);
            WaitForMultipleObjects(threadCount, threads, TRUE, INFINITE);

            // Kill threads
            for (int t = 0; t < threadCount; t++)
                CloseHandle(threads[t]);

            // Report!
            printf("threads=%d ", threadCount);
            printf("lockInterval=%e ", benchmarkParams[b].lockInterval);
            printf("lockDuration=%f ", (float) s / steps);
            ThreadStats totals = { 0 };
            for (int t = 0; t < threadCount; t++)
            {
                totals.workDone += globalState.threadStats[t].workDone;
                totals.iterations += globalState.threadStats[t].iterations;
                totals.overshoot += globalState.threadStats[t].overshoot;
            }
            printf("workDone=%I64d ", totals.workDone);
            printf("iterations=%I64d ", totals.iterations);
            printf("overshoot=%I64d ", totals.overshoot);
            puts("");
        }
    }

    // Shutdown
    CloseHandle(globalState.startEvent);
    CloseHandle(globalState.initSemaphore);
    DeleteCriticalSection(&globalState.criticalSection);

	return 0;
}

